//
// module data
// © 10.11.2007 Ingo D. Rullhusen
//

#include <stdio.h>
#include <float.h>

#include <iostream>
#include <string>

#include "module_header.h"

#include "natconst.h"
#include "module_data.h"
#include "tetraquad.h"
#include "triquad.h"

using namespace std;

int MainData_version      = MainData_version_;
int FIELD_version         = FIELD_version_;
int module_header_version = module_header_version_;

int version = 1;
char *name  = "Dielectric Module";


extern "C" void* new_data(){ return( (void*) new ModuleData ); }

extern "C" void* dpl_data( void *di_ ){ return( (void*) new ModuleData( (ModuleData*)di_ ) ); }

extern "C" void del_data( void *di_ ){ delete (ModuleData*)di_; }


const int matrix_para_no = 9;
const char *matrix_para[ matrix_para_no ] =
{
   "mesh unit / m",
   "center point x / m",
   "center point y / m",
   "center point z / m",
   "theta / °",
   "phi / °",
   "psi / °",
   "msh file",
   "no. of physical entities"
};


const int material_para_no = 3;
const char *material_para[ material_para_no ] =
{
   "Gmsh phys. entity no.",
   "conductivity / S/m",
   "ε_r"
};



extern "C" void matrix_get_parameter( const MainData *MD, void *data_,
                                      string **param, string **value, int *no )
{
   ModuleData *data = (ModuleData*)data_;
   const int buffsize = 100;
   char sb[buffsize];

   *no = matrix_para_no + data->no_material*material_para_no;
   *param = new string[*no];
   *value = new string[*no];

   int id = 0;

   sprintf(sb,"%g",data->unit);
   (*param)[id] = matrix_para[id];
   (*value)[id] = sb;
   id++;

   sprintf(sb,"%g",data->cpx);
   (*param)[id] = matrix_para[id];
   (*value)[id] = sb;
   id++;

   sprintf(sb,"%g",data->cpy);
   (*param)[id] = matrix_para[id];
   (*value)[id] = sb;
   id++;

   sprintf(sb,"%g",data->cpz);
   (*param)[id] = matrix_para[id];
   (*value)[id] = sb;
   id++;

   sprintf(sb,"%g", data->theta * 180.0 / M_PI );
   (*param)[id] = matrix_para[id];
   (*value)[id] = sb;
   id++;

   sprintf(sb,"%g", data->phi * 180.0 / M_PI );
   (*param)[id] = matrix_para[id];
   (*value)[id] = sb;
   id++;

   sprintf(sb,"%g", data->psi * 180.0 / M_PI );  
   (*param)[id] = matrix_para[id];
   (*value)[id] = sb;
   id++;


   (*param)[id] = matrix_para[id];
   (*value)[id] = data->msh_file_name;
   id++;



   sprintf(sb,"%i", data->no_material);
   (*param)[id] = matrix_para[id];
   (*value)[id] = sb;
   id++;

   int id2 = 0;
   for ( int m=0; m<data->no_material; m++ )
   {
      id2 = 0;
      snprintf( sb, buffsize, "%i:%s", m+1, material_para[id2] );
      (*param)[id] = sb;
      snprintf( sb, buffsize, "%i", data->ml[m].phys_entity );
      (*value)[id] = sb;
      id2++;
      id++;

      snprintf( sb, buffsize, "%i:%s", m+1, material_para[id2] );
      (*param)[id] = sb;
      snprintf( sb, buffsize, "%g", data->ml[m].conductivity );
      (*value)[id] = sb;
      id2++;
      id++;

      snprintf( sb, buffsize, "%i:%s", m+1, material_para[id2] );
      (*param)[id] = sb;
      snprintf( sb, buffsize, "%g", data->ml[m].epsilon_r );
      (*value)[id] = sb;
      id2++;
      id++;
   }

}



extern "C" void matrix_set_parameter( const MainData *MD, void *data_,
                                      const string *param, const string *value, int no )
{
   ModuleData *data = (ModuleData*)data_;
   const int buffsize = 100;
   char sb[buffsize];
   int j, nm;

   nm = data->no_material;

   for ( int i=0; i<no; i++ )
   {
      for ( j=0; j<matrix_para_no; j++ ) if ( matrix_para[j] == param[i] ) break;

      switch(j)
      {
         case 0: { data->unit = atof( value[i].c_str() ); break; }
         case 1: { data->cpx = atof( value[i].c_str() ); break; } 
         case 2: { data->cpy = atof( value[i].c_str() ); break; } 
         case 3: { data->cpz = atof( value[i].c_str() ); break; } 
         case 4: { data->theta = atof( value[i].c_str() )*M_PI/180.0; break; }
         case 5: { data->phi   = atof( value[i].c_str() )*M_PI/180.0; break; }
         case 6: { data->psi   = atof( value[i].c_str() )*M_PI/180.0; break; }
         case 7: { data->msh_file_name = value[i]; break; }
         case 8: { nm = atoi( value[i].c_str() ); break; }
      }
   }

// alloc & copy old
   if ( nm != data->no_material )
   {
      MATERIAL *mt = new MATERIAL[ nm ];

      for ( int i=0; i<( nm<data->no_material ? nm : data->no_material ); i++ )
      {
         mt[i] = data->ml[i];
      }

      delete [] data->ml;
      data->ml = mt;
      mt = NULL;

      data->no_material = nm;
   }

// set values
   for ( int i=0; i<no; i++ )
   {
      for ( int m=0; m<data->no_material; m++ )
      {
         for ( j=0; j<material_para_no; j++ )
         {
            snprintf( sb, buffsize, "%i:%s", m+1, material_para[j] );

            if ( sb == param[i] )
            {
               switch(j)
               {
                  case 0: { data->ml[m].phys_entity  = atoi( value[i].c_str() ); break; }
                  case 1: { data->ml[m].conductivity = atof( value[i].c_str() ); break; }
                  case 2: { data->ml[m].epsilon_r    = atof( value[i].c_str() ); break; }
               }
            }

         }
      }
   }

}



extern "C" int matrix_geo_setup( const MainData *MD, void *data_,
                                 string **omsg, int *nomsg )
{
   ModuleData *data = (ModuleData*)data_;

   data->setup_geo();


// Green's function
   delete data->grf;
   data->grf = new GREEN( MD->k0 );

// statistic
   double min, max, l;

   min = DBL_MAX;
   max = -1.0;

   for ( TETRAEDER_LIST::iterator it = data->tetralist.begin();
         it != data->tetralist.end();
         ++it )
   {

      l = HYPOT( data->plist[ it->p1 ].x - data->plist[ it->p2 ].x,
                 data->plist[ it->p1 ].y - data->plist[ it->p2 ].y,
                 data->plist[ it->p1 ].z - data->plist[ it->p2 ].z  );
      if ( l < min ) min = l;
      if ( l > max ) max = l;

      l = HYPOT( data->plist[ it->p1 ].x - data->plist[ it->p3 ].x,
                 data->plist[ it->p1 ].y - data->plist[ it->p3 ].y,   
                 data->plist[ it->p1 ].z - data->plist[ it->p3 ].z  );
      if ( l < min ) min = l;
      if ( l > max ) max = l;

      l = HYPOT( data->plist[ it->p1 ].x - data->plist[ it->p4 ].x,
                 data->plist[ it->p1 ].y - data->plist[ it->p4 ].y,   
                 data->plist[ it->p1 ].z - data->plist[ it->p4 ].z  );
      if ( l < min ) min = l;
      if ( l > max ) max = l;

      l = HYPOT( data->plist[ it->p2 ].x - data->plist[ it->p4 ].x,
                 data->plist[ it->p2 ].y - data->plist[ it->p4 ].y,
                 data->plist[ it->p2 ].z - data->plist[ it->p4 ].z  );
      if ( l < min ) min = l;
      if ( l > max ) max = l;

      l = HYPOT( data->plist[ it->p3 ].x - data->plist[ it->p4 ].x,
                 data->plist[ it->p3 ].y - data->plist[ it->p4 ].y,
                 data->plist[ it->p3 ].z - data->plist[ it->p4 ].z  );
      if ( l < min ) min = l;
      if ( l > max ) max = l;

      l = HYPOT( data->plist[ it->p3 ].x - data->plist[ it->p2 ].x,
                 data->plist[ it->p3 ].y - data->plist[ it->p2 ].y,
                 data->plist[ it->p3 ].z - data->plist[ it->p2 ].z  );
      if ( l < min ) min = l;
      if ( l > max ) max = l;

   }

// message
   *nomsg = 3;
   *omsg = new string[*nomsg];

   const int buffsize = 64;
   char buff[buffsize];
   snprintf( buff, buffsize, "edge length (min/max): %g m / %g m", min, max );
   (*omsg)[0] = buff;
   snprintf( buff, buffsize, "no. of points   : %u", (unsigned)data->plist.size() );
   (*omsg)[1] = buff;
   snprintf( buff, buffsize, "no. of basis fct: %u", (unsigned)data->bfktlist.size() );
   (*omsg)[2] = buff;


   return( data->bfktlist.size() );
}



extern "C" void matrix_renumber( const MainData *MD, void *data_, int *perm, int *rperm )
{
   ModuleData *data = (ModuleData*)data_;

   for ( TETRAEDER_LIST::iterator it = data->tetralist.begin();
         it != data->tetralist.end();
         ++it )
   {
      if ( it->sign.sp1 ) it->Tb1 = perm[ it->Tb1 ];
      if ( it->sign.sp2 ) it->Tb2 = perm[ it->Tb2 ];
      if ( it->sign.sp3 ) it->Tb3 = perm[ it->Tb3 ];
      if ( it->sign.sp4 ) it->Tb4 = perm[ it->Tb4 ];
   }

   Tbfkt_LIST tbt;
   tbt.clear();

   for ( unsigned int i=0; i<data->bfktlist.size(); ++i )
   {
      tbt.push_back( data->bfktlist[ rperm[i] ] );
   }

   data->bfktlist.swap(tbt);
   tbt.clear();
}


extern "C" void matrix_bounding_box( const MainData *MD, void *data_,
                                     double *cpos_x, double *cpos_y, double *cpos_z, 
                                     double *size_x, double *size_y, double *size_z )
{
   ModuleData *data = (ModuleData*)data_;

   if ( data->plist.size() < 1 )
   {
      *cpos_x = MD->cpos_x;
      *cpos_y = MD->cpos_y;
      *cpos_z = MD->cpos_z;

      *size_x = *size_y = *size_z = 0.0;

      return;
   }

   double minx, maxx, miny, maxy, minz, maxz;
   minx = maxx = data->plist[0].x;
   miny = maxy = data->plist[0].y;
   minz = maxz = data->plist[0].z;

   for ( POINT_LIST::iterator it = data->plist.begin()+1;
         it != data->plist.end();
         it++ )
   {
      minx = ( minx < it->x ? minx : it->x );
      miny = ( miny < it->y ? miny : it->y );
      minz = ( minz < it->z ? minz : it->z );

      maxx = ( maxx > it->x ? maxx : it->x );
      maxy = ( maxy > it->y ? maxy : it->y );
      maxz = ( maxz > it->z ? maxz : it->z );
   }

   *size_x = maxx - minx;
   *size_y = maxy - miny;
   *size_z = maxz - minz;

   *cpos_x = 0.5 * ( maxx + minx );
   *cpos_y = 0.5 * ( maxy + miny );
   *cpos_z = 0.5 * ( maxz + minz );
}


#ifdef HAVE_LIBGL
extern "C" void matrix_add_to_OGLlist( const MainData *MD, void *data_, GLuint list )
{
   ModuleData *data = (ModuleData*)data_;

   data->ogl_triangles( list );
}
#endif


extern "C"  void matrix_bfkt_bounding_box( const MainData *MD, void *data_, int nbfkt,
                                           int *sdno,
                                           double **cpos_x, double **cpos_y, double **cpos_z,
                                           double **size_x, double **size_y, double **size_z  )
{
   ModuleData *data = (ModuleData*)data_;

   *sdno = 0;

   *cpos_x = NULL;
   *cpos_y = NULL;
   *cpos_z = NULL;

   *size_x = NULL;
   *size_y = NULL;
   *size_z = NULL;

   if ( nbfkt < 0 || nbfkt > (int)data->bfktlist.size() )
      throw( "invalid basis function number" );

   double minx, maxx, miny, maxy, minz, maxz,
          minx2, maxx2, miny2, maxy2, minz2, maxz2;
   POINT *P;

   *sdno = 2;

   *cpos_x = new double[*sdno];
   *cpos_y = new double[*sdno];
   *cpos_z = new double[*sdno];

   *size_x = new double[*sdno];
   *size_y = new double[*sdno];
   *size_z = new double[*sdno];


   P = &data->plist[ data->bfktlist[nbfkt].trip1 ];
   minx = maxx = P->x;
   miny = maxy = P->y;
   minz = maxz = P->z;

   P = &data->plist[ data->bfktlist[nbfkt].trip2 ];
   minx = ( minx < P->x ? minx : P->x );
   miny = ( miny < P->y ? miny : P->y );
   minz = ( minz < P->z ? minz : P->z );

   maxx = ( maxx > P->x ? maxx : P->x );
   maxy = ( maxy > P->y ? maxy : P->y );
   maxz = ( maxz > P->z ? maxz : P->z );

   P = &data->plist[ data->bfktlist[nbfkt].trip3 ];
   minx = ( minx < P->x ? minx : P->x );
   miny = ( miny < P->y ? miny : P->y );
   minz = ( minz < P->z ? minz : P->z );

   maxx = ( maxx > P->x ? maxx : P->x );
   maxy = ( maxy > P->y ? maxy : P->y );
   maxz = ( maxz > P->z ? maxz : P->z );

   P = &data->plist[ data->bfktlist[nbfkt].sp1 ];
   minx2 = ( minx < P->x ? minx : P->x );
   miny2 = ( miny < P->y ? miny : P->y );
   minz2 = ( minz < P->z ? minz : P->z );

   maxx2 = ( maxx > P->x ? maxx : P->x );
   maxy2 = ( maxy > P->y ? maxy : P->y );
   maxz2 = ( maxz > P->z ? maxz : P->z );

   (*size_x)[0] = maxx2 - minx2;
   (*size_y)[0] = maxy2 - miny2;
   (*size_z)[0] = maxz2 - minz2;

   (*cpos_x)[0] = 0.5 * ( maxx2 + minx2 );
   (*cpos_y)[0] = 0.5 * ( maxy2 + miny2 );
   (*cpos_z)[0] = 0.5 * ( maxz2 + minz2 );

   if ( data->bfktlist[nbfkt].sp2 > -1 )
   {
      P = &data->plist[ data->bfktlist[nbfkt].sp2 ];
      minx2 = ( minx < P->x ? minx : P->x );
      miny2 = ( miny < P->y ? miny : P->y );
      minz2 = ( minz < P->z ? minz : P->z );

      maxx2 = ( maxx > P->x ? maxx : P->x );
      maxy2 = ( maxy > P->y ? maxy : P->y );
      maxz2 = ( maxz > P->z ? maxz : P->z );
   }
   else
   {
      minx2 = minx;
      miny2 = miny;
      minz2 = minz;

      maxx2 = maxx;
      maxy2 = maxy;
      maxz2 = maxz;
   }

   (*size_x)[1] = maxx2 - minx2;
   (*size_y)[1] = maxy2 - miny2;
   (*size_z)[1] = maxz2 - minz2;

   (*cpos_x)[1] = 0.5 * ( maxx2 + minx2 );
   (*cpos_y)[1] = 0.5 * ( maxy2 + miny2 );
   (*cpos_z)[1] = 0.5 * ( maxz2 + minz2 );
}


extern "C" void matrix_bfkt_moment( const MainData *MD, void *data_,
                                    int nbfkt, int sdno,
                                    double cpos_x, double cpos_y, double cpos_z,
                                    double len_x, double len_y, double len_z,
                                    int nx, int ny, int nz,
                                    double *mx, double *my, double *mz, double *md )
{
   ModuleData *data = (ModuleData*)data_;
   static const triquad t_irule(7);
   static const tetraquad irule(5);
   int t;
   double ipx, ipy, ipz, dx, dy, dz, powpow;

   *mx = *my = *mz = *md = 0.0;

   if ( sdno == 0 )
   {

// + tetraeder
   t = data->bfktlist[ nbfkt ].tetra1;
   for ( int i=0; i<irule.no; i++ )
   {

      ipx =   irule.lambda1[i] * data->plist[ data->tetralist[t].p1 ].x   
            + irule.lambda2[i] * data->plist[ data->tetralist[t].p2 ].x 
            + irule.lambda3[i] * data->plist[ data->tetralist[t].p3 ].x
            + irule.lambda4[i] * data->plist[ data->tetralist[t].p4 ].x;

      ipy =   irule.lambda1[i] * data->plist[ data->tetralist[t].p1 ].y   
            + irule.lambda2[i] * data->plist[ data->tetralist[t].p2 ].y 
            + irule.lambda3[i] * data->plist[ data->tetralist[t].p3 ].y
            + irule.lambda4[i] * data->plist[ data->tetralist[t].p4 ].y;

      ipz =   irule.lambda1[i] * data->plist[ data->tetralist[t].p1 ].z   
            + irule.lambda2[i] * data->plist[ data->tetralist[t].p2 ].z 
            + irule.lambda3[i] * data->plist[ data->tetralist[t].p3 ].z
            + irule.lambda4[i] * data->plist[ data->tetralist[t].p4 ].z;

      dx = ( ipx - cpos_x ) / len_x;
      dy = ( ipy - cpos_y ) / len_y;
      dz = ( ipz - cpos_z ) / len_z;

      powpow =  pow( dx, nx ) * pow( dy, ny ) * pow( dz, nz );

// S(chaubert)WG functions
      *mx += irule.weights[i] * powpow
             * ( ipx - data->plist[ data->bfktlist[nbfkt].sp1 ].x ) / 3.0;
      *my += irule.weights[i] * powpow
             * ( ipy - data->plist[ data->bfktlist[nbfkt].sp1 ].y ) / 3.0;
      *mz += irule.weights[i] * powpow
             * ( ipz - data->plist[ data->bfktlist[nbfkt].sp1 ].z ) / 3.0;

      *md += irule.weights[i] * powpow;
// V^+ / V^+
   }

   }
   else if ( sdno == 1 )
   {

   t = data->bfktlist[ nbfkt ].tetra2;
   if ( t > -1 )
   {

// - tetraeder
   for ( int i=0; i<irule.no; i++ )
   {

      ipx =   irule.lambda1[i] * data->plist[ data->tetralist[t].p1 ].x   
            + irule.lambda2[i] * data->plist[ data->tetralist[t].p2 ].x 
            + irule.lambda3[i] * data->plist[ data->tetralist[t].p3 ].x
            + irule.lambda4[i] * data->plist[ data->tetralist[t].p4 ].x;

      ipy =   irule.lambda1[i] * data->plist[ data->tetralist[t].p1 ].y   
            + irule.lambda2[i] * data->plist[ data->tetralist[t].p2 ].y 
            + irule.lambda3[i] * data->plist[ data->tetralist[t].p3 ].y
            + irule.lambda4[i] * data->plist[ data->tetralist[t].p4 ].y;

      ipz =   irule.lambda1[i] * data->plist[ data->tetralist[t].p1 ].z   
            + irule.lambda2[i] * data->plist[ data->tetralist[t].p2 ].z 
            + irule.lambda3[i] * data->plist[ data->tetralist[t].p3 ].z
            + irule.lambda4[i] * data->plist[ data->tetralist[t].p4 ].z;

      dx = ( ipx - cpos_x ) / len_x;
      dy = ( ipy - cpos_y ) / len_y;
      dz = ( ipz - cpos_z ) / len_z;

      powpow =  pow( dx, nx ) * pow( dy, ny ) * pow( dz, nz );

// S(chaubert)WG functions
      *mx -= irule.weights[i] * powpow
             * ( ipx - data->plist[ data->bfktlist[nbfkt].sp2 ].x ) / 3.0;
      *my -= irule.weights[i] * powpow
             * ( ipy - data->plist[ data->bfktlist[nbfkt].sp2 ].y ) / 3.0;
      *mz -= irule.weights[i] * powpow
             * ( ipz - data->plist[ data->bfktlist[nbfkt].sp2 ].z ) / 3.0;

      *md -= irule.weights[i] * powpow;
// V^- / V^-
   }

   }
   else
   {

// triangle
   for ( int i=0; i<t_irule.no; i++ )
   {

      ipx =   t_irule.lambda1[i] * data->plist[ data->bfktlist[nbfkt].trip1 ].x
            + t_irule.lambda2[i] * data->plist[ data->bfktlist[nbfkt].trip2 ].x
            + t_irule.lambda3[i] * data->plist[ data->bfktlist[nbfkt].trip3 ].x;

      ipy =   t_irule.lambda1[i] * data->plist[ data->bfktlist[nbfkt].trip1 ].y
            + t_irule.lambda2[i] * data->plist[ data->bfktlist[nbfkt].trip2 ].y
            + t_irule.lambda3[i] * data->plist[ data->bfktlist[nbfkt].trip3 ].y;

      ipz =   t_irule.lambda1[i] * data->plist[ data->bfktlist[nbfkt].trip1 ].z
            + t_irule.lambda2[i] * data->plist[ data->bfktlist[nbfkt].trip2 ].z
            + t_irule.lambda3[i] * data->plist[ data->bfktlist[nbfkt].trip3 ].z;

      dx = ( ipx - cpos_x ) / len_x;
      dy = ( ipy - cpos_y ) / len_y;
      dz = ( ipz - cpos_z ) / len_z;

      powpow =  pow( dx, nx ) * pow( dy, ny ) * pow( dz, nz );

// S(chaubert)WG functions

      *md -= t_irule.weights[i] * powpow;
// A / A
   }

   }   // end: if tetra/tri

   }
   else
   {
      throw( "invalid subdomain number" );
   }

}


/***
extern "C" void matrix_ele_setup( const MainData *MD, void *data,
                                  string **omsg, int *nomsg )
{ *omsg = NULL; *nomsg = 0; }
***/



void add_potential_tetra( complex<double> A[3], complex<double> &phi,
                          double signum,
                          ModuleData *data, int qtetra, POINT SP,
                                            int atetra, POINT ap )
{
   const complex<double> cONE(1.0,0.0);

   static const tetraquad iI(4);

   bool singu;
   double rv2[3], volume;
   complex<double> Gd[3][3], Gs, v[3], J[3], div, J0[3], div0;
   POINT qp;

   if ( data->tetraeder_neigh(atetra,qtetra) )
   {
      singu = true;
      if ( atetra == qtetra )
      {
         J0[0] = ( ap.x - SP.x ) / 3.0;
         J0[1] = ( ap.y - SP.y ) / 3.0;
         J0[2] = ( ap.z - SP.z ) / 3.0;
         div0 = cONE;
      }
      else
      {
         data->tetraeder_center( qtetra, qp );

         J0[0] = ( qp.x - SP.x ) / 3.0;
         J0[1] = ( qp.y - SP.y ) / 3.0;
         J0[2] = ( qp.z - SP.z ) / 3.0;
         div0 = cONE;
      }
   }
   else
      singu = false;

   for ( int ip2=0; ip2<iI.no; ++ip2 )
   {
      qp.x = iI.lambda1[ip2] * data->plist[ data->tetralist[qtetra].p1 ].x
             + iI.lambda2[ip2] * data->plist[ data->tetralist[qtetra].p2 ].x
             + iI.lambda3[ip2] * data->plist[ data->tetralist[qtetra].p3 ].x
             + iI.lambda4[ip2] * data->plist[ data->tetralist[qtetra].p4 ].x;
      qp.y = iI.lambda1[ip2] * data->plist[ data->tetralist[qtetra].p1 ].y   
             + iI.lambda2[ip2] * data->plist[ data->tetralist[qtetra].p2 ].y 
             + iI.lambda3[ip2] * data->plist[ data->tetralist[qtetra].p3 ].y
             + iI.lambda4[ip2] * data->plist[ data->tetralist[qtetra].p4 ].y;
      qp.z = iI.lambda1[ip2] * data->plist[ data->tetralist[qtetra].p1 ].z   
             + iI.lambda2[ip2] * data->plist[ data->tetralist[qtetra].p2 ].z 
             + iI.lambda3[ip2] * data->plist[ data->tetralist[qtetra].p3 ].z
             + iI.lambda4[ip2] * data->plist[ data->tetralist[qtetra].p4 ].z;

      rv2[0] = ( qp.x - SP.x ) / 3.0;
      rv2[1] = ( qp.y - SP.y ) / 3.0;
      rv2[2] = ( qp.z - SP.z ) / 3.0;

      div = cONE;

      if ( singu )
      {
         J[0] = rv2[0];
         J[1] = rv2[1];
         J[2] = rv2[2];
         data->grf->EP_green_ex( v, Gs, J, div, J0, div0, ap, qp );

         A[0] += signum * iI.weights[ip2] * v[0];  // /volume
         A[1] += signum * iI.weights[ip2] * v[1];  // /volume
         A[2] += signum * iI.weights[ip2] * v[2];  // /volume

         phi  += signum * iI.weights[ip2] * Gs;  // /volume
      }
      else
      {   
         data->grf->EP_green( Gd, Gs, ap, qp );

         A[0] += signum * iI.weights[ip2] *
                 ( Gd[0][0]*rv2[0] + Gd[0][1]*rv2[1] + Gd[0][2]*rv2[2] );  // /volume
         A[1] += signum * iI.weights[ip2] *
                 ( Gd[1][0]*rv2[0] + Gd[1][1]*rv2[1] + Gd[1][2]*rv2[2] );  // /volume
         A[2] += signum * iI.weights[ip2] *
                 ( Gd[2][0]*rv2[0] + Gd[2][1]*rv2[1] + Gd[2][2]*rv2[2] );  // /volume

         phi  += signum * iI.weights[ip2] * Gs * div;  // /volume
      }
   }

   if ( singu )
   {
      volume = data->tetraeder_volume(qtetra);
      J0[0] /= volume;
      J0[1] /= volume;
      J0[2] /= volume;
      div0  /= volume;

      data->grf->EP_green_re( v, div, J0, div0,
                              data->plist[ data->tetralist[qtetra].p1 ],
                                 data->plist[ data->tetralist[qtetra].p2 ],
                                    data->plist[ data->tetralist[qtetra].p3 ],
                                       data->plist[ data->tetralist[qtetra].p4 ], ap );
      A[0] += signum * v[0];
      A[1] += signum * v[1];
      A[2] += signum * v[2];
      phi  += signum * div;  
   }

}

void add_potential_tri( complex<double> &phi,
                        ModuleData *data, Tbfkt qBFKT, int atetra, POINT ap )
{
   const complex<double> cNULL(0.0,0.0);
   const complex<double> cONE(1.0,0.0);

   static const triquad iI(3);

   bool singu;
   complex<double> Gd[3][3], Gs, v[3], J[3], div, J0[3], div0;
   POINT qp;

   J0[0] = J0[1] = J0[2] = cNULL;
   J[0] = J[1] = J[2] = cNULL;

   if ( data->tetraeder_neigh(atetra,qBFKT.tetra1) )
   {
      singu = true;
      div0 = -cONE;
   }
   else
      singu = false;

   for ( int ip2=0; ip2<iI.no; ++ip2 )
   {
      qp.x = iI.lambda1[ip2] * data->plist[ qBFKT.trip1 ].x
             + iI.lambda2[ip2] * data->plist[ qBFKT.trip2 ].x
             + iI.lambda3[ip2] * data->plist[ qBFKT.trip3 ].x;
      qp.y = iI.lambda1[ip2] * data->plist[ qBFKT.trip1 ].y
             + iI.lambda2[ip2] * data->plist[ qBFKT.trip2 ].y
             + iI.lambda3[ip2] * data->plist[ qBFKT.trip3 ].y;
      qp.z = iI.lambda1[ip2] * data->plist[ qBFKT.trip1 ].z
             + iI.lambda2[ip2] * data->plist[ qBFKT.trip2 ].z
             + iI.lambda3[ip2] * data->plist[ qBFKT.trip3 ].z;

      div = -cONE;

      if ( singu )
      {
         data->grf->EP_green_ex( v, Gs, J, div, J0, div0, ap, qp );

         phi += iI.weights[ip2] * Gs;  // /area
      }
      else
      {
         data->grf->EP_green( Gd, Gs, ap, qp );

         phi += iI.weights[ip2] * Gs * div;  // /area
      }
   }

   if ( singu )
   {
      div0 /= qBFKT.triangle_area( data->plist );

      data->grf->EP_green_re( v, div, J0, div0,
                              data->plist[ qBFKT.trip1 ],
                                 data->plist[ qBFKT.trip2 ],
                                    data->plist[ qBFKT.trip3 ], ap );
      phi += div;
   }

}



extern "C" complex<double> matrix_ele( const MainData *MD, void *data_, int e1, int e2 )
{
   ModuleData *data = (ModuleData*)data_;

   const complex<double> cNULL(0.0,0.0); 
   const complex<double> cONE(1.0,0.0);
   const complex<double> cJ(0.0,1.0);

   static const triquad t_oI(1);
   static const tetraquad oI(1);

   int atetra, qtetra;
   double rv1[3];
   complex<double> zval, A[3], phi, signum;
   POINT ap;

#ifdef BOUNDARY_CHECK
   if( e1 < 0 || e1 >= (int)data->bfktlist.size() ||
       e2 < 0 || e2 >= (int)data->bfktlist.size()    )
   {
      cerr << "Error in DielectricModule::matrix_ele(): line " << __LINE__ << "!" << endl;
      exit(1);
   }
#endif

   complex<double> jomega = cJ*MD->omega;

   zval = cNULL;

//
// f_i f_j
   double l1_1[3], l1_2[3], l1_3[3], l2_1[3], l2_2[3], l2_3[3], fac[3][3];
   POINT *sP, *rP1, *rP2, *rP3;

   if ( data->bfktlist[e1].tetra1 == data->bfktlist[e2].tetra1 ||
        data->bfktlist[e1].tetra1 == data->bfktlist[e2].tetra2    )
   {
      atetra = data->bfktlist[e1].tetra1;

      signum = cONE;
      sP = &data->plist[ data->bfktlist[e1].sp1 ];
      rP1 = &data->plist[ data->bfktlist[e1].trip1 ];
      rP2 = &data->plist[ data->bfktlist[e1].trip2 ];
      rP3 = &data->plist[ data->bfktlist[e1].trip3 ];

      l1_1[0] = rP1->x - sP->x;
      l1_1[1] = rP1->y - sP->y;
      l1_1[2] = rP1->z - sP->z;

      l1_2[0] = rP2->x - sP->x;
      l1_2[1] = rP2->y - sP->y;
      l1_2[2] = rP2->z - sP->z;

      l1_3[0] = rP3->x - sP->x;
      l1_3[1] = rP3->y - sP->y;
      l1_3[2] = rP3->z - sP->z;

      if ( data->bfktlist[e2].tetra1 == atetra )
      {
         signum *= +1.0;
         sP = &data->plist[ data->bfktlist[e2].sp1 ];   
         rP1 = &data->plist[ data->bfktlist[e2].trip1 ];
         rP2 = &data->plist[ data->bfktlist[e2].trip2 ];
         rP3 = &data->plist[ data->bfktlist[e2].trip3 ];
      }
      else if ( data->bfktlist[e2].tetra2 == atetra )
      {
         signum *= -1.0;
         sP = &data->plist[ data->bfktlist[e2].sp2 ];
         rP1 = &data->plist[ data->bfktlist[e2].trip1 ];
         rP2 = &data->plist[ data->bfktlist[e2].trip2 ];
         rP3 = &data->plist[ data->bfktlist[e2].trip3 ];
      }
      else
      {
         cerr << "Error in DielectricModule matrix_ele(): internal line "
              << __LINE__ << "!" << endl;
         exit(1);
      }

      l2_1[0] = rP1->x - sP->x;
      l2_1[1] = rP1->y - sP->y;
      l2_1[2] = rP1->z - sP->z;

      l2_2[0] = rP2->x - sP->x;
      l2_2[1] = rP2->y - sP->y;
      l2_2[2] = rP2->z - sP->z;

      l2_3[0] = rP3->x - sP->x;
      l2_3[1] = rP3->y - sP->y;
      l2_3[2] = rP3->z - sP->z;


      fac[0][0] = ( data->bfktlist[e1].trip1 == data->bfktlist[e2].trip1 ? 0.1 : 0.05 );
      fac[0][1] = ( data->bfktlist[e1].trip1 == data->bfktlist[e2].trip2 ? 0.1 : 0.05 );
      fac[0][2] = ( data->bfktlist[e1].trip1 == data->bfktlist[e2].trip3 ? 0.1 : 0.05 );
      fac[1][0] = ( data->bfktlist[e1].trip2 == data->bfktlist[e2].trip1 ? 0.1 : 0.05 );
      fac[1][1] = ( data->bfktlist[e1].trip2 == data->bfktlist[e2].trip2 ? 0.1 : 0.05 );
      fac[1][2] = ( data->bfktlist[e1].trip2 == data->bfktlist[e2].trip3 ? 0.1 : 0.05 );
      fac[2][0] = ( data->bfktlist[e1].trip3 == data->bfktlist[e2].trip1 ? 0.1 : 0.05 );
      fac[2][1] = ( data->bfktlist[e1].trip3 == data->bfktlist[e2].trip2 ? 0.1 : 0.05 );
      fac[2][2] = ( data->bfktlist[e1].trip3 == data->bfktlist[e2].trip3 ? 0.1 : 0.05 );


      signum /= ( data->ml[ data->tetralist[atetra].material ].conductivity
                  + jomega * EPS0 * ( data->ml[ data->tetralist[atetra].material ].epsilon_r - 1.0 ) )
                * 9.0 * data->tetraeder_volume(atetra);   // V / V^2


      zval -= signum * (

               ( l1_1[0]*l2_1[0] + l1_1[1]*l2_1[1] + l1_1[2]*l2_1[2] ) * fac[0][0]
              +( l1_1[0]*l2_2[0] + l1_1[1]*l2_2[1] + l1_1[2]*l2_2[2] ) * fac[0][1]
              +( l1_1[0]*l2_3[0] + l1_1[1]*l2_3[1] + l1_1[2]*l2_3[2] ) * fac[0][2]

              +( l1_2[0]*l2_1[0] + l1_2[1]*l2_1[1] + l1_2[2]*l2_1[2] ) * fac[1][0]
              +( l1_2[0]*l2_2[0] + l1_2[1]*l2_2[1] + l1_2[2]*l2_2[2] ) * fac[1][1]
              +( l1_2[0]*l2_3[0] + l1_2[1]*l2_3[1] + l1_2[2]*l2_3[2] ) * fac[1][2]

              +( l1_3[0]*l2_1[0] + l1_3[1]*l2_1[1] + l1_3[2]*l2_1[2] ) * fac[2][0]
              +( l1_3[0]*l2_2[0] + l1_3[1]*l2_2[1] + l1_3[2]*l2_2[2] ) * fac[2][1]
              +( l1_3[0]*l2_3[0] + l1_3[1]*l2_3[1] + l1_3[2]*l2_3[2] ) * fac[2][2]

                       );

   }

   if ( ( data->bfktlist[e1].tetra2 == data->bfktlist[e2].tetra1 ||
          data->bfktlist[e1].tetra2 == data->bfktlist[e2].tetra2    )
        && data->bfktlist[e1].tetra2 > -1                             )
   {
      atetra = data->bfktlist[e1].tetra2;

      signum = -cONE;
      sP = &data->plist[ data->bfktlist[e1].sp2 ];
      rP1 = &data->plist[ data->bfktlist[e1].trip1 ];
      rP2 = &data->plist[ data->bfktlist[e1].trip2 ];
      rP3 = &data->plist[ data->bfktlist[e1].trip3 ];

      l1_1[0] = rP1->x - sP->x;
      l1_1[1] = rP1->y - sP->y;
      l1_1[2] = rP1->z - sP->z;

      l1_2[0] = rP2->x - sP->x;
      l1_2[1] = rP2->y - sP->y;
      l1_2[2] = rP2->z - sP->z;

      l1_3[0] = rP3->x - sP->x;
      l1_3[1] = rP3->y - sP->y;
      l1_3[2] = rP3->z - sP->z;

      if ( data->bfktlist[e2].tetra1 == atetra )
      {
         signum *= +1.0;
         sP = &data->plist[ data->bfktlist[e2].sp1 ];   
         rP1 = &data->plist[ data->bfktlist[e2].trip1 ];
         rP2 = &data->plist[ data->bfktlist[e2].trip2 ];
         rP3 = &data->plist[ data->bfktlist[e2].trip3 ];
      }
      else if ( data->bfktlist[e2].tetra2 == atetra )
      {
         signum *= -1.0;
         sP = &data->plist[ data->bfktlist[e2].sp2 ];
         rP1 = &data->plist[ data->bfktlist[e2].trip1 ];
         rP2 = &data->plist[ data->bfktlist[e2].trip2 ];
         rP3 = &data->plist[ data->bfktlist[e2].trip3 ];
      }
      else
      {
         cerr << "Error in DielectricModule matrix_ele(): internal line "
              << __LINE__ << "!" << endl;
         exit(1);
      }

      l2_1[0] = rP1->x - sP->x;
      l2_1[1] = rP1->y - sP->y;
      l2_1[2] = rP1->z - sP->z;

      l2_2[0] = rP2->x - sP->x;
      l2_2[1] = rP2->y - sP->y;
      l2_2[2] = rP2->z - sP->z;

      l2_3[0] = rP3->x - sP->x;
      l2_3[1] = rP3->y - sP->y;
      l2_3[2] = rP3->z - sP->z;


      fac[0][0] = ( data->bfktlist[e1].trip1 == data->bfktlist[e2].trip1 ? 0.1 : 0.05 );
      fac[0][1] = ( data->bfktlist[e1].trip1 == data->bfktlist[e2].trip2 ? 0.1 : 0.05 );
      fac[0][2] = ( data->bfktlist[e1].trip1 == data->bfktlist[e2].trip3 ? 0.1 : 0.05 );
      fac[1][0] = ( data->bfktlist[e1].trip2 == data->bfktlist[e2].trip1 ? 0.1 : 0.05 );
      fac[1][1] = ( data->bfktlist[e1].trip2 == data->bfktlist[e2].trip2 ? 0.1 : 0.05 );
      fac[1][2] = ( data->bfktlist[e1].trip2 == data->bfktlist[e2].trip3 ? 0.1 : 0.05 );
      fac[2][0] = ( data->bfktlist[e1].trip3 == data->bfktlist[e2].trip1 ? 0.1 : 0.05 );
      fac[2][1] = ( data->bfktlist[e1].trip3 == data->bfktlist[e2].trip2 ? 0.1 : 0.05 );
      fac[2][2] = ( data->bfktlist[e1].trip3 == data->bfktlist[e2].trip3 ? 0.1 : 0.05 );


      signum /= ( data->ml[ data->tetralist[atetra].material ].conductivity
                  + jomega * EPS0 * ( data->ml[ data->tetralist[atetra].material ].epsilon_r - 1.0 ) )
                * 9.0 * data->tetraeder_volume(atetra);   // V / V^2


      zval -= signum * (

               ( l1_1[0]*l2_1[0] + l1_1[1]*l2_1[1] + l1_1[2]*l2_1[2] ) * fac[0][0]
              +( l1_1[0]*l2_2[0] + l1_1[1]*l2_2[1] + l1_1[2]*l2_2[2] ) * fac[0][1]
              +( l1_1[0]*l2_3[0] + l1_1[1]*l2_3[1] + l1_1[2]*l2_3[2] ) * fac[0][2]

              +( l1_2[0]*l2_1[0] + l1_2[1]*l2_1[1] + l1_2[2]*l2_1[2] ) * fac[1][0]
              +( l1_2[0]*l2_2[0] + l1_2[1]*l2_2[1] + l1_2[2]*l2_2[2] ) * fac[1][1]
              +( l1_2[0]*l2_3[0] + l1_2[1]*l2_3[1] + l1_2[2]*l2_3[2] ) * fac[1][2]

              +( l1_3[0]*l2_1[0] + l1_3[1]*l2_1[1] + l1_3[2]*l2_1[2] ) * fac[2][0]
              +( l1_3[0]*l2_2[0] + l1_3[1]*l2_2[1] + l1_3[2]*l2_2[2] ) * fac[2][1]
              +( l1_3[0]*l2_3[0] + l1_3[1]*l2_3[1] + l1_3[2]*l2_3[2] ) * fac[2][2]

                       );

   }


//
// E(J)

// tetraeder 1 p
   atetra = data->bfktlist[e1].tetra1;

   for ( int ip1=0; ip1<oI.no; ip1++ )
   {
      ap.x = oI.lambda1[ip1] * data->plist[ data->tetralist[atetra].p1 ].x
             + oI.lambda2[ip1] * data->plist[ data->tetralist[atetra].p2 ].x
             + oI.lambda3[ip1] * data->plist[ data->tetralist[atetra].p3 ].x
             + oI.lambda4[ip1] * data->plist[ data->tetralist[atetra].p4 ].x;
      ap.y = oI.lambda1[ip1] * data->plist[ data->tetralist[atetra].p1 ].y   
             + oI.lambda2[ip1] * data->plist[ data->tetralist[atetra].p2 ].y 
             + oI.lambda3[ip1] * data->plist[ data->tetralist[atetra].p3 ].y
             + oI.lambda4[ip1] * data->plist[ data->tetralist[atetra].p4 ].y;
      ap.z = oI.lambda1[ip1] * data->plist[ data->tetralist[atetra].p1 ].z   
             + oI.lambda2[ip1] * data->plist[ data->tetralist[atetra].p2 ].z 
             + oI.lambda3[ip1] * data->plist[ data->tetralist[atetra].p3 ].z
             + oI.lambda4[ip1] * data->plist[ data->tetralist[atetra].p4 ].z;

      A[0] = cNULL;
      A[1] = cNULL;
      A[2] = cNULL;
      phi  = cNULL;


// tetraeder 2 p
      qtetra = data->bfktlist[e2].tetra1;
      add_potential_tetra( A, phi, +1.0,
                          data, qtetra, data->plist[ data->bfktlist[e2].sp1 ],
                                atetra, ap );


      qtetra = data->bfktlist[e2].tetra2;
      if ( qtetra < 0 )
      {
// triangle 2
         add_potential_tri( phi, data, data->bfktlist[e2], atetra, ap );
      }
      else
      {
// tetraeder 2 m
         add_potential_tetra( A, phi, -1.0,
                              data, qtetra, data->plist[ data->bfktlist[e2].sp2 ],
                                    atetra, ap );
      }


      rv1[0] = ( ap.x - data->plist[ data->bfktlist[e1].sp1 ].x ) / 3.0;
      rv1[1] = ( ap.y - data->plist[ data->bfktlist[e1].sp1 ].y ) / 3.0;
      rv1[2] = ( ap.z - data->plist[ data->bfktlist[e1].sp1 ].z ) / 3.0;

      zval += oI.weights[ip1] * (
                 - jomega * ( rv1[0]*A[0] + rv1[1]*A[1] + rv1[2]*A[2] )
                 - 1.0 * phi / jomega
                                 );

   }

   atetra = data->bfktlist[e1].tetra2;
   if ( atetra < 0 )
   {
// triangle 1

   for ( int ip1=0; ip1<t_oI.no; ip1++ )
   {
      ap.x = t_oI.lambda1[ip1] * data->plist[ data->bfktlist[e1].trip1 ].x
             + t_oI.lambda2[ip1] * data->plist[ data->bfktlist[e1].trip2 ].x
             + t_oI.lambda3[ip1] * data->plist[ data->bfktlist[e1].trip3 ].x;
      ap.y = t_oI.lambda1[ip1] * data->plist[ data->bfktlist[e1].trip1 ].y   
             + t_oI.lambda2[ip1] * data->plist[ data->bfktlist[e1].trip2 ].y 
             + t_oI.lambda3[ip1] * data->plist[ data->bfktlist[e1].trip3 ].y;
      ap.z = t_oI.lambda1[ip1] * data->plist[ data->bfktlist[e1].trip1 ].z   
             + t_oI.lambda2[ip1] * data->plist[ data->bfktlist[e1].trip2 ].z 
             + t_oI.lambda3[ip1] * data->plist[ data->bfktlist[e1].trip3 ].z;

      A[0] = cNULL;
      A[1] = cNULL;
      A[2] = cNULL;
      phi  = cNULL;


// tetraeder 2 p
      qtetra = data->bfktlist[e2].tetra1;
      add_potential_tetra( A, phi, +1.0,
                          data, qtetra, data->plist[ data->bfktlist[e2].sp1 ],
                                data->bfktlist[e1].tetra1, ap );


      qtetra = data->bfktlist[e2].tetra2;
      if ( qtetra < 0 )
      {
// triangle 2
         add_potential_tri( phi, data, data->bfktlist[e2],
                                 data->bfktlist[e1].tetra1, ap );
      }
      else
      {
// tetraeder 2 m
         add_potential_tetra( A, phi, -1.0,
                              data, qtetra, data->plist[ data->bfktlist[e2].sp2 ],
                                    data->bfktlist[e1].tetra1, ap );
      }


      zval -= t_oI.weights[ip1] * ( -1.0 *  phi / jomega );

   }

   }
   else
   {
// tetraeder 1 m

   for ( int ip1=0; ip1<oI.no; ip1++ )
   {
      ap.x = oI.lambda1[ip1] * data->plist[ data->tetralist[atetra].p1 ].x
             + oI.lambda2[ip1] * data->plist[ data->tetralist[atetra].p2 ].x
             + oI.lambda3[ip1] * data->plist[ data->tetralist[atetra].p3 ].x
             + oI.lambda4[ip1] * data->plist[ data->tetralist[atetra].p4 ].x;
      ap.y = oI.lambda1[ip1] * data->plist[ data->tetralist[atetra].p1 ].y   
             + oI.lambda2[ip1] * data->plist[ data->tetralist[atetra].p2 ].y 
             + oI.lambda3[ip1] * data->plist[ data->tetralist[atetra].p3 ].y
             + oI.lambda4[ip1] * data->plist[ data->tetralist[atetra].p4 ].y;
      ap.z = oI.lambda1[ip1] * data->plist[ data->tetralist[atetra].p1 ].z   
             + oI.lambda2[ip1] * data->plist[ data->tetralist[atetra].p2 ].z 
             + oI.lambda3[ip1] * data->plist[ data->tetralist[atetra].p3 ].z
             + oI.lambda4[ip1] * data->plist[ data->tetralist[atetra].p4 ].z;

      A[0] = cNULL;
      A[1] = cNULL;
      A[2] = cNULL;
      phi  = cNULL;


// tetraeder 2 p
      qtetra = data->bfktlist[e2].tetra1;
      add_potential_tetra( A, phi, +1.0,
                          data, qtetra, data->plist[ data->bfktlist[e2].sp1 ],
                                atetra, ap );


      qtetra = data->bfktlist[e2].tetra2;
      if ( qtetra < 0 )
      {
// triangle 2
         add_potential_tri( phi, data, data->bfktlist[e2], atetra, ap );
      }
      else
      {
// tetraeder 2 m
         add_potential_tetra( A, phi, -1.0,
                              data, qtetra, data->plist[ data->bfktlist[e2].sp2 ],
                                    atetra, ap );
      }


      rv1[0] = ( data->plist[ data->bfktlist[e1].sp2 ].x - ap.x ) / 3.0;
      rv1[1] = ( data->plist[ data->bfktlist[e1].sp2 ].y - ap.y ) / 3.0;
      rv1[2] = ( data->plist[ data->bfktlist[e1].sp2 ].z - ap.z ) / 3.0;

      zval += oI.weights[ip1] * (
                 - jomega * ( rv1[0]*A[0] + rv1[1]*A[1] + rv1[2]*A[2] )
                 - (-1.0) * phi / jomega
                                 );

   }

   }   // end: if ( atetra < 0 )

   return(zval);
}

///extern "C" void matrix_ele_cleanup( const MainData *MD, void *data ){}




const int extract_para_no = 1;
const char *extract_para[ extract_para_no ] =
{ 
   "no. of probes"
};

const int P_para_no = 3;
const char *P_para[ P_para_no ] =
{
   "pos x / m",
   "pos y / m",
   "pos z / m"
};

extern "C" void extract_get_parameter( const MainData *MD, void *data_,
                                       string **param, string **value,
                                       int *no )
{
   ModuleData *data = (ModuleData*)data_;
   const int buffsize = 100;
   char sb[buffsize];

   *no = extract_para_no + data->probe_plist.size()*P_para_no;
   *param = new string[*no];
   *value = new string[*no];
   int id = 0;

   sprintf(sb,"%u",(unsigned)data->probe_plist.size());
   (*param)[id] = extract_para[0];
   (*value)[id] = sb;
   id++;

   int id2;
   for ( unsigned int w=0; w<data->probe_plist.size(); w++ )
   {
      id2 = 0;
      snprintf( sb, buffsize, "%i:%s", w+1, P_para[id2] );
      (*param)[id] = sb;
      snprintf( sb, buffsize, "%g", data->probe_plist[w].x );
      (*value)[id] = sb;
      id2++;
      id++;

      snprintf( sb, buffsize, "%i:%s", w+1, P_para[id2] );
      (*param)[id] = sb;
      snprintf( sb, buffsize, "%g", data->probe_plist[w].y );
      (*value)[id] = sb;
      id2++;
      id++;

      snprintf( sb, buffsize, "%i:%s", w+1, P_para[id2] );
      (*param)[id] = sb;
      snprintf( sb, buffsize, "%g", data->probe_plist[w].z );
      (*value)[id] = sb;
      id2++;
      id++;
   }

}

extern "C" void extract_set_parameter( const MainData *MD, void *data_,
                                       const string *param, const string *value,
                                       int no )
{
   ModuleData *data = (ModuleData*)data_;
   const int buffsize = 100;
   char sb[buffsize];
   int j, ng;

   ng = data->probe_plist.size();

   for ( int i=0; i<no; ++i )
   {
      for ( j=0; j<extract_para_no; ++j ) if ( extract_para[j] == param[i] ) break;

      switch(j)
      {
         case 0: { ng = atoi( value[i].c_str() ); break; }
      }
   }

// probe
   if ( ng != (int)data->probe_plist.size() )
   {
      data->probe_plist.resize(ng);
   }

   for ( int i=0; i<no; i++ )   
   {
      for ( unsigned int g=0; g<data->probe_plist.size(); ++g )
      {

         for ( j=0; j<P_para_no; j++ )
         {
            snprintf( sb, buffsize, "%i:%s", g+1, P_para[j] );

            if ( sb == param[i] )
            {
               switch(j)
               {
                  case 0: { data->probe_plist[g].x = atof( value[i].c_str() ); break; }
                  case 1: { data->probe_plist[g].y = atof( value[i].c_str() ); break; }
                  case 2: { data->probe_plist[g].z = atof( value[i].c_str() ); break; }
               }
            }

         }

      }
   }

}

extern "C" void extract_add_to_OGLlist( const MainData *MD, void *data_, GLuint gl_list )
{
   ModuleData *data = (ModuleData*)data_;
   const GLfloat clr_1[4] = { 1.00, 0.00, 1.00, 1.0 };
   const GLfloat clr_2[4] = { 0.50, 0.50, 0.50, 1.0 };
   int t;
   double R;
   TETRAEDER *tetra;

   R = 0.5*sqrt( MD->size_x*MD->size_x
                 + MD->size_y*MD->size_y
                 + MD->size_z*MD->size_z );

   for ( POINT_LIST::iterator P = data->probe_plist.begin();
                              P != data->probe_plist.end();
                              ++P )
   {
// point
      glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, clr_1 );

      glBegin(GL_LINES);

      glVertex3d( P->x - 0.1*R, P->y, P->z );
      glVertex3d( P->x + 0.1*R, P->y, P->z );

      glVertex3d( P->x, P->y - 0.1*R, P->z );
      glVertex3d( P->x, P->y + 0.1*R, P->z );

      glVertex3d( P->x, P->y, P->z - 0.1*R );
      glVertex3d( P->x, P->y, P->z + 0.1*R );

      glEnd();

// tetraeder
      t = data->in_tetraeder(*P);
      if ( t > -1 )
      {
         tetra = &data->tetralist[t];
         glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, clr_2 );

         glBegin(GL_LINES);

         glVertex3d( data->plist[ tetra->p1 ].x,
                     data->plist[ tetra->p1 ].y,
                     data->plist[ tetra->p1 ].z  );
         glVertex3d( data->plist[ tetra->p2 ].x,
                     data->plist[ tetra->p2 ].y,
                     data->plist[ tetra->p2 ].z  );

         glVertex3d( data->plist[ tetra->p2 ].x,
                     data->plist[ tetra->p2 ].y,
                     data->plist[ tetra->p2 ].z  );
         glVertex3d( data->plist[ tetra->p3 ].x,
                     data->plist[ tetra->p3 ].y,
                     data->plist[ tetra->p3 ].z  );

         glVertex3d( data->plist[ tetra->p3 ].x,
                     data->plist[ tetra->p3 ].y,
                     data->plist[ tetra->p3 ].z  );
         glVertex3d( data->plist[ tetra->p1 ].x,
                     data->plist[ tetra->p1 ].y,
                     data->plist[ tetra->p1 ].z  );

         glVertex3d( data->plist[ tetra->p1 ].x,
                     data->plist[ tetra->p1 ].y,
                     data->plist[ tetra->p1 ].z  );
         glVertex3d( data->plist[ tetra->p4 ].x,
                     data->plist[ tetra->p4 ].y,
                     data->plist[ tetra->p4 ].z  );

         glVertex3d( data->plist[ tetra->p2 ].x,
                     data->plist[ tetra->p2 ].y,
                     data->plist[ tetra->p2 ].z  );
         glVertex3d( data->plist[ tetra->p4 ].x,
                     data->plist[ tetra->p4 ].y,
                     data->plist[ tetra->p4 ].z  );

         glVertex3d( data->plist[ tetra->p3 ].x,
                     data->plist[ tetra->p3 ].y,
                     data->plist[ tetra->p3 ].z  );
         glVertex3d( data->plist[ tetra->p4 ].x,
                     data->plist[ tetra->p4 ].y,
                     data->plist[ tetra->p4 ].z  );

         glEnd();
      }

   }

}

extern "C" void extract_run( const MainData *MD, void *data_, bool dummy,
                             complex<double> *JV, int noJ, const FIELD *f,
                             string **omsg, int *nomsg )
{
   ModuleData *data = (ModuleData*)data_;
   const int buffsize = 100;
   int t;
   char buff[buffsize];
   double V;
   complex<double> J[3], fac;
   TETRAEDER *tetra;
   vector<string> lines;

   lines.clear();

   for ( POINT_LIST::iterator P = data->probe_plist.begin();
                              P != data->probe_plist.end();
                              ++P )
   {
      snprintf( buff, buffsize, "probe at (%g;%g;%g):\n", P->x, P->y, P->z );
      lines.push_back(buff);

// tetraeder
      t = data->in_tetraeder(*P);
      if ( t < 0 )
      {
         snprintf( buff, buffsize, "Probe outside dielectric body!\n" );
         lines.push_back(buff);
         continue;
      }

      tetra = &data->tetralist[ t ];
      V = data->tetraeder_volume( t );

// current
      J[0] = J[1] = J[2] = complex<double>(0.0,0.0);
// bfkt 1
      if ( tetra->sign.sp1 != 0 )
      {
         fac = (double)tetra->sign.sp1 * JV[ tetra->Tb1 ] / (3.0*V);

         J[0] += fac * ( P->x - data->plist[ tetra->p1 ].x );
         J[1] += fac * ( P->y - data->plist[ tetra->p1 ].y );
         J[2] += fac * ( P->z - data->plist[ tetra->p1 ].z );
      }
// bfkt 2
      if ( tetra->sign.sp2 != 0 )
      {
         fac = (double)tetra->sign.sp2 * JV[ tetra->Tb2 ] / (3.0*V);

         J[0] += fac * ( P->x - data->plist[ tetra->p2 ].x );
         J[1] += fac * ( P->y - data->plist[ tetra->p2 ].y );
         J[2] += fac * ( P->z - data->plist[ tetra->p2 ].z );
      }
// bfkt 3
      if ( tetra->sign.sp3 != 0 )
      {
         fac = (double)tetra->sign.sp3 * JV[ tetra->Tb3 ] / (3.0*V);

         J[0] += fac * ( P->x - data->plist[ tetra->p3 ].x );
         J[1] += fac * ( P->y - data->plist[ tetra->p3 ].y );
         J[2] += fac * ( P->z - data->plist[ tetra->p3 ].z );
      }
// bfkt 4
      if ( tetra->sign.sp4 != 0 )
      {
         fac = (double)tetra->sign.sp4 * JV[ tetra->Tb4 ] / (3.0*V);

         J[0] += fac * ( P->x - data->plist[ tetra->p4 ].x );
         J[1] += fac * ( P->y - data->plist[ tetra->p4 ].y );
         J[2] += fac * ( P->z - data->plist[ tetra->p4 ].z );
      }

// output
// J
      snprintf( buff, buffsize, " abs(J_x) = %g A/m^2 \t arg(J_x) = %g°",
                abs(J[0]), arg(J[0])*180.0/M_PI );
      lines.push_back(buff);
      snprintf( buff, buffsize, " abs(J_y) = %g A/m^2 \t arg(J_y) = %g°",
                abs(J[1]), arg(J[1])*180.0/M_PI );
      lines.push_back(buff);
      snprintf( buff, buffsize, " abs(J_z) = %g A/m^2 \t arg(J_z) = %g°",
                abs(J[2]), arg(J[2])*180.0/M_PI );
      lines.push_back(buff);
      lines.push_back("");
// E
      fac = 1.0 / ( data->ml[ tetra->material ].conductivity
                    + complex<double>(0.0,MD->omega)
                      * EPS0 * ( data->ml[ tetra->material ].epsilon_r - 1.0 ) );
      J[0] *= fac;
      J[1] *= fac;
      J[2] *= fac;
      snprintf( buff, buffsize, " abs(E_x) = %g V/m \t arg(E_x) = %g°",
                abs(J[0]), arg(J[0])*180.0/M_PI );
      lines.push_back(buff);
      snprintf( buff, buffsize, " abs(E_y) = %g V/m \t arg(E_y) = %g°",
                abs(J[1]), arg(J[1])*180.0/M_PI );
      lines.push_back(buff);
      snprintf( buff, buffsize, " abs(E_z) = %g V/m \t arg(E_z) = %g°",
                abs(J[2]), arg(J[2])*180.0/M_PI );
      lines.push_back(buff);
      lines.push_back("");
   }

//
// generate messages
   *nomsg = lines.size();
   *omsg = new string[ *nomsg ];

   for ( int g=0; g<*nomsg; g++ )
   {
      (*omsg)[g] = lines[g];
   }

   lines.clear();
}
