//
// module data
// (c) 7.7.2007 Ingo D. Rullhusen
//

#ifndef MODULE_DATA_H
#define MODULE_DATA_H

#include <GL/gl.h>

#include <string>
#include <complex>

#include "point.h"
#include "green.h"

using namespace std;


class MATERIAL
{
   public:
      int phys_entity;
      double conductivity, epsilon_r;
};


class TETRAEDER
{
   public:
      int p1, p2, p3, p4;
      int Tb1, Tb2, Tb3, Tb4;

      struct
      {
         signed short sp1:2;  // -1; 0; +1
         signed short sp2:2;  // -1; 0; +1
         signed short sp3:2;  // -1; 0; +1
         signed short sp4:2;  // -1; 0; +1
      } sign;

      int material;

      TETRAEDER();
};

typedef vector<TETRAEDER> TETRAEDER_LIST;


class Tbfkt
{
   public:
      int tetra1, tetra2;
      int trip1, trip2, trip3;
      int sp1, sp2;

      Tbfkt();

      double triangle_area( POINT_LIST plist ) const;
};

typedef vector<Tbfkt> Tbfkt_LIST;


class ModuleData
{
   public:

// matrix
      double unit;
      double cpx, cpy, cpz;
      double theta, phi, psi;

      string msh_file_name;

      int no_material;
      MATERIAL *ml;


      POINT_LIST plist;

      TETRAEDER_LIST tetralist;

      Tbfkt_LIST bfktlist;


      GREEN *grf;


// extraction
      POINT_LIST probe_plist;


// functions
      ModuleData();
      ModuleData( const ModuleData *d );
      ~ModuleData();

      void setup_geo();

#ifdef HAVE_LIBGL
      void ogl_triangles( GLuint list );
#endif

      void tetraeder_center( int tno, POINT &p ) const;
      bool tetraeder_neigh( int tno1, int tno2 ) const;
      double tetraeder_volume( int tno ) const;
      int in_tetraeder( POINT &p ) const;   // <0: out of all

};


inline double HYPOT( double x, double y, double z ){ return(sqrt(x*x+y*y+z*z)); }


#endif
