//
// Class for tetrahedral quadrature
// (c) 16.2.2007 Ingo D. Rullhusen
//

#include <math.h>
#include <iostream>

#include "tetraquad.h"


tetraquad::tetraquad( int n )
{
   double r;

   no = n;
   weights = new double[n];
   lambda1 = new double[n];
   lambda2 = new double[n];
   lambda3 = new double[n];
   lambda4 = new double[n];

   switch (n)
   {
      case 1:
      {
         weights[0] = 1.0;
         lambda1[0] = 1.0 / 4.0;
         lambda2[0] = 1.0 / 4.0;
         lambda3[0] = 1.0 / 4.0;
         lambda4[0] = 1.0 / 4.0;
         break;
      }
      case 4:
      {
         weights[0] = weights[1] = weights[2] = weights[3] = 1.0 / 4.0;

         r = 1.0 / sqrt( 5.0 );

         lambda1[0] = r + ( 1.0 - r ) / 4.0;
         lambda2[0] = ( 1.0 - r ) / 4.0;
         lambda3[0] = ( 1.0 - r ) / 4.0;
         lambda4[0] = ( 1.0 - r ) / 4.0;

         lambda1[1] = ( 1.0 - r ) / 4.0;         
         lambda2[1] = r + ( 1.0 - r ) / 4.0;
         lambda3[1] = ( 1.0 - r ) / 4.0;
         lambda4[1] = ( 1.0 - r ) / 4.0;

         lambda1[2] = ( 1.0 - r ) / 4.0;             
         lambda2[2] = ( 1.0 - r ) / 4.0;
         lambda3[2] = r + ( 1.0 - r ) / 4.0;
         lambda4[2] = ( 1.0 - r ) / 4.0;

         lambda1[3] = ( 1.0 - r ) / 4.0;             
         lambda2[3] = ( 1.0 - r ) / 4.0;    
         lambda3[3] = ( 1.0 - r ) / 4.0;
         lambda4[3] = r + ( 1.0 - r ) / 4.0;
         break;
      }
      case 5:
      {
         weights[0] = weights[1] = weights[2] = weights[3] = 9.0 / 20.0;
         weights[4] = - 4.0 / 5.0;

         lambda1[0] = 3.0 / 6.0;
         lambda2[0] = 1.0 / 6.0;
         lambda3[0] = 1.0 / 6.0;
         lambda4[0] = 1.0 / 6.0;

         lambda1[1] = 1.0 / 6.0;
         lambda2[1] = 3.0 / 6.0;
         lambda3[1] = 1.0 / 6.0;
         lambda4[1] = 1.0 / 6.0;

         lambda1[2] = 1.0 / 6.0;
         lambda2[2] = 1.0 / 6.0;
         lambda3[2] = 3.0 / 6.0;
         lambda4[2] = 1.0 / 6.0;

         lambda1[3] = 1.0 / 6.0;
         lambda2[3] = 1.0 / 6.0;
         lambda3[3] = 1.0 / 6.0;
         lambda4[3] = 3.0 / 6.0;

         lambda1[4] = 1.0 / 4.0;
         lambda2[4] = 1.0 / 4.0;
         lambda3[4] = 1.0 / 4.0;
         lambda4[4] = 1.0 / 4.0;
         break;
      }
      default:
      {
         cerr << "Error in tetraquad::tetraquad(no): no = " << n << " not supported!" << endl;
         exit(1);
      }
   }
}


tetraquad::~tetraquad()
{
   no = 0;
   delete [] weights;
   delete [] lambda1;
   delete [] lambda2;
   delete [] lambda3;
   delete [] lambda4;
}
