//
// Class for triangular quadrature
// (c) 16.4.2003 Ingo D. Rullhusen
//

#include <stdlib.h>
#include <iostream>
#include <math.h>
#include "triquad.h"

triquad::triquad( int n )
{
   int  i;
   double a, b;

   no = n;
   weights = new double[n];
   lambda1 = new double[n];
   lambda2 = new double[n];
   lambda3 = new double[n];

   switch (n)
   {
      case 1:
      {
         weights[0] = 1.0;
         lambda1[0] = 1.0 / 3.0;
         lambda2[0] = 1.0 / 3.0;
         lambda3[0] = 1.0 / 3.0;
         break;
      }
      case 3:
      {
         weights[0] = 1.0 / 3.0;
         weights[1] = 1.0 / 3.0;
         weights[2] = 1.0 / 3.0;

         lambda1[0] = 2.0 / 3.0;
         lambda2[0] = 1.0 / 6.0;
         lambda3[0] = 1.0 / 6.0;

         lambda1[1] = 1.0 / 6.0;
         lambda2[1] = 2.0 / 3.0;
         lambda3[1] = 1.0 / 6.0;

         lambda1[2] = 1.0 / 6.0;
         lambda2[2] = 1.0 / 6.0;
         lambda3[2] = 2.0 / 3.0;
         break;
      }
      case 7:
      {
         weights[0] = ( 155.0 - sqrt(15.0) ) / 1200.0;
         weights[1] = ( 155.0 - sqrt(15.0) ) / 1200.0;
         weights[2] = ( 155.0 - sqrt(15.0) ) / 1200.0;
         weights[3] = ( 155.0 + sqrt(15.0) ) / 1200.0;
         weights[4] = ( 155.0 + sqrt(15.0) ) / 1200.0;
         weights[5] = ( 155.0 + sqrt(15.0) ) / 1200.0;
         weights[6] = 9.0 / 40.0;

         lambda1[0] = ( 1.0 + sqrt(15.0) ) / 7.0
                      + ( 6.0 - sqrt(15.0) ) / ( 7.0 * 3.0 );
         lambda2[0] = ( 6.0 - sqrt(15.0) ) / ( 7.0 * 3.0 );
         lambda3[0] = ( 6.0 - sqrt(15.0) ) / ( 7.0 * 3.0 );

         lambda1[1] = ( 6.0 - sqrt(15.0) ) / ( 7.0 * 3.0 );
         lambda2[1] = ( 1.0 + sqrt(15.0) ) / 7.0
                      + ( 6.0 - sqrt(15.0) ) / ( 7.0 * 3.0 );
         lambda3[1] = ( 6.0 - sqrt(15.0) ) / ( 7.0 * 3.0 );

         lambda1[2] = ( 6.0 - sqrt(15.0) ) / ( 7.0 * 3.0 );
         lambda2[2] = ( 6.0 - sqrt(15.0) ) / ( 7.0 * 3.0 );
         lambda3[2] = ( 1.0 + sqrt(15.0) ) / 7.0
                      + ( 6.0 - sqrt(15.0) ) / ( 7.0 * 3.0 );

         lambda1[3] = ( 1.0 - sqrt(15.0) ) / 7.0
                      + ( 6.0 + sqrt(15.0) ) / ( 7.0 * 3.0 );
         lambda2[3] = ( 6.0 + sqrt(15.0) ) / ( 7.0 * 3.0 );
         lambda3[3] = ( 6.0 + sqrt(15.0) ) / ( 7.0 * 3.0 );

         lambda1[4] = ( 6.0 + sqrt(15.0) ) / ( 7.0 * 3.0 );
         lambda2[4] = ( 1.0 - sqrt(15.0) ) / 7.0
                      + ( 6.0 + sqrt(15.0) ) / ( 7.0 * 3.0 );
         lambda3[4] = ( 6.0 + sqrt(15.0) ) / ( 7.0 * 3.0 );

         lambda1[5] = ( 6.0 + sqrt(15.0) ) / ( 7.0 * 3.0 );
         lambda2[5] = ( 6.0 + sqrt(15.0) ) / ( 7.0 * 3.0 );
         lambda3[5] = ( 1.0 - sqrt(15.0) ) / 7.0
                      + ( 6.0 + sqrt(15.0) ) / ( 7.0 * 3.0 );

         lambda1[6] = 1.0 / 3.0;
         lambda2[6] = 1.0 / 3.0;
         lambda3[6] = 1.0 / 3.0;
         break;
      }
      case 21:
      {
// part 1
         weights[0] = ( 155.0 - sqrt(15.0) ) / ( 1200.0 * 3.0 );
         weights[1] = weights[0];
         weights[2] = weights[0];
         weights[3] = ( 155.0 + sqrt(15.0) ) / ( 1200.0 * 3.0 );
         weights[4] = weights[3];
         weights[5] = weights[3];
         weights[6] = 9.0 / ( 40.0 * 3.0 );

         a = ( 1.0 + sqrt(15.0) ) / 7.0;
         b = ( 6.0 - sqrt(15.0) ) / 7.0;
         lambda1[0] = a + b / 3.0 + b / 9.0;
         lambda2[0] = b / 3.0 + b / 9.0;
         lambda3[0] = b / 9.0;

         lambda1[1] = b / 3.0 + b / 9.0;
         lambda2[1] = a + b / 3.0 + b / 9.0;
         lambda3[1] = b / 9.0;

         lambda1[2] = a / 3.0 + b / 3.0 + b / 9.0;
         lambda2[2] = a / 3.0 + b / 3.0 + b / 9.0;
         lambda3[2] = a / 3.0 + b / 9.0;

         a = ( 1.0 - sqrt(15.0) ) / 7.0;
         b = ( 6.0 + sqrt(15.0) ) / 7.0;
         lambda1[3] = a + b / 3.0 + b / 9.0;
         lambda2[3] = b / 3.0 + b / 9.0;
         lambda3[3] = b / 9.0;

         lambda1[4] = b / 3.0 + b / 9.0;
         lambda2[4] = a + b / 3.0 + b / 9.0;
         lambda3[4] = b / 9.0;

         lambda1[5] = a / 3.0 + b / 3.0 + b / 9.0;
         lambda2[5] = a / 3.0 + b / 3.0 + b / 9.0;
         lambda3[5] = a / 3.0 + b / 9.0;

         lambda1[6] = 4.0 / 9.0;
         lambda2[6] = 4.0 / 9.0;
         lambda3[6] = 1.0 / 9.0;

// part 2
         for ( i=0; i<7; i++ ) weights[7+i] = weights[i];

         a = ( 1.0 + sqrt(15.0) ) / 7.0;
         b = ( 6.0 - sqrt(15.0) ) / 7.0;
         lambda3[7+0] = a + b / 3.0 + b / 9.0;
         lambda1[7+0] = b / 3.0 + b / 9.0;
         lambda2[7+0] = b / 9.0;

         lambda3[7+1] = b / 3.0 + b / 9.0;
         lambda1[7+1] = a + b / 3.0 + b / 9.0;
         lambda2[7+1] = b / 9.0;

         lambda3[7+2] = a / 3.0 + b / 3.0 + b / 9.0;
         lambda1[7+2] = a / 3.0 + b / 3.0 + b / 9.0;
         lambda2[7+2] = a / 3.0 + b / 9.0;

         a = ( 1.0 - sqrt(15.0) ) / 7.0;
         b = ( 6.0 + sqrt(15.0) ) / 7.0;
         lambda3[7+3] = a + b / 3.0 + b / 9.0;
         lambda1[7+3] = b / 3.0 + b / 9.0;
         lambda2[7+3] = b / 9.0;

         lambda3[7+4] = b / 3.0 + b / 9.0;
         lambda1[7+4] = a + b / 3.0 + b / 9.0;
         lambda2[7+4] = b / 9.0;

         lambda3[7+5] = a / 3.0 + b / 3.0 + b / 9.0;
         lambda1[7+5] = a / 3.0 + b / 3.0 + b / 9.0;
         lambda2[7+5] = a / 3.0 + b / 9.0;

         lambda3[7+6] = 4.0 / 9.0;
         lambda1[7+6] = 4.0 / 9.0;
         lambda2[7+6] = 1.0 / 9.0;

// part 3
         for ( i=0; i<7; i++ ) weights[14+i] = weights[i];

         a = ( 1.0 + sqrt(15.0) ) / 7.0;
         b = ( 6.0 - sqrt(15.0) ) / 7.0;
         lambda2[14+0] = a + b / 3.0 + b / 9.0;
         lambda3[14+0] = b / 3.0 + b / 9.0;
         lambda1[14+0] = b / 9.0;

         lambda2[14+1] = b / 3.0 + b / 9.0;
         lambda3[14+1] = a + b / 3.0 + b / 9.0;
         lambda1[14+1] = b / 9.0;

         lambda2[14+2] = a / 3.0 + b / 3.0 + b / 9.0;
         lambda3[14+2] = a / 3.0 + b / 3.0 + b / 9.0;
         lambda1[14+2] = a / 3.0 + b / 9.0;

         a = ( 1.0 - sqrt(15.0) ) / 7.0;
         b = ( 6.0 + sqrt(15.0) ) / 7.0;
         lambda2[14+3] = a + b / 3.0 + b / 9.0;
         lambda3[14+3] = b / 3.0 + b / 9.0;
         lambda1[14+3] = b / 9.0;

         lambda2[14+4] = b / 3.0 + b / 9.0;
         lambda3[14+4] = a + b / 3.0 + b / 9.0;
         lambda1[14+4] = b / 9.0;

         lambda2[14+5] = a / 3.0 + b / 3.0 + b / 9.0;
         lambda3[14+5] = a / 3.0 + b / 3.0 + b / 9.0;
         lambda1[14+5] = a / 3.0 + b / 9.0;

         lambda2[14+6] = 4.0 / 9.0;
         lambda3[14+6] = 4.0 / 9.0;
         lambda1[14+6] = 1.0 / 9.0;
         break;
      }
      default:
      {
         cerr << "Error in triquad::triquad(no): no = " << n << " not supported!" << endl;
         exit(1);
      }
   }
}

triquad::~triquad()
{
   no = 0;
   delete [] weights;
   delete [] lambda1;
   delete [] lambda2;
   delete [] lambda3;
}
